// ==UserScript==
// @name         IMAX mobile
// @namespace    maid.aural.inspection
// @version      1
// @description  Firefox137向け：IMAX mobile
// @author       gemini
// @match        https://*youtube.com/*
// @grant        none
// ==/UserScript==

(async function() {
    'use strict';

    


    // --- 設定変数 ---
    const TAP_AUDIO_DURATION_SEC = 3.00; // 基本周波数推定のためのデータ収集時間（秒）
    const TARGET_SAMPLE_RATE = 192000; // 目標サンプルレート (Hz)

    // メディア要素待機設定
    const MEDIA_WAIT_INTERVAL_MS = 300;
    const MEDIA_WAIT_MAX_TIME_SEC = 20;

    // LocalEnhance 処理用設定 (Worklet内で使用) - v7.0/v8.0から維持
    const LE_CHUNK_SIZE_WORKLET = 128;
    const LE_RMS_THRESHOLD_WORKLET = 0.01;

    // 「弾性波動」関連定数 (Worklet内で使用) - v7.0/v8.0から維持
    const EW_VELOCITY = 340.29;
    const EW_WAVELENGTH_FACTOR = EW_VELOCITY / 440;
    const EW_ENHANCEMENT_FACTOR = Math.sin(Math.PI / EW_WAVELENGTH_FACTOR) * 0.8;

    // 「Gamma式補完」関連定数 (Worklet内で使用 - 再設計版、補間強化)
    const GAMMA_L = 1.00;
    const GAMMA_T = 1.00;
    const GAMMA_MU0 = 1.00;
    const GAMMA_KAPPA = 1.00;
    const GAMMA_ALPHAP = 1.00;
    const GAMMA_THRESHOLD_WORKLET = 0.3;
    const GAMMA_INTERPOLATION_FACTOR = 0.5; // 補間値をどの程度混ぜるか (0.0から1.0)
     const GAMMA_ENHANCE_GAIN = 1.2; // 補間適用時の追加ゲイン (強調用)


    // --- 音響物理学応用設定 ---
    const RHO0 = 1.2;
    const C = 343;
    const HORN_L = 0.5;
    const HORN_M = 2.5;
    const HORN_S0 = 0.01;
    const HORN_CUTOFF_FREQ_CALCULATED = HORN_M * C / (4 * Math.PI);
    const TUBE_LENGTH = 0.5;
    const TUBE_RESONANCE_FREQ = (1 * C) / (2 * TUBE_LENGTH);


     // --- IMAX再現強化設定 (v8.0ベースから維持) ---

     // サブハーモニック合成設定
    const SUBHARMONIC_RATIO = 2;
    const SUBHARMONIC_GAIN = 0.6;

     // Mid/Side 処理設定
    const MID_GAIN_DB = -2;
    const SIDE_GAIN_DB = 7;

     // 簡易空間反射模倣設定 (Convolutionを主とするため、簡易反射は補助)
     const EARLY_REFLECTION_DELAYS_MS = [5, 10, 15];
     const EARLY_REFLECTION_GAINS = [0.2, 0.15, 0.1];

     // Convolution Reverb 設定
     const CONVOLVER_IR_DURATION_SEC = 0.2;
     const CONVOLVER_IR_DECAY_RATE = 5.0;
     const CONVOLVER_IR_NUM_REFLECTIONS = 5;
     const CONVOLVER_WET_GAIN = 0.4;
     const CONVOLVER_DRY_GAIN = 1.0;


     // EQ設定調整 - v8.0ベース維持
     const ULTIMATE_IMAX_EQ_SETTINGS_ADVANCED = [
        { type: 'lowshelf', freq: 20, gain: 15 },
        { type: 'peaking', freq: 50, gain: 6, Q: 1.8 },
        { type: 'peaking', freq: 100, gain: 3, Q: 1.5 },
        { type: 'peaking', freq: 200, gain: -3, Q: 1.0 },
        { type: 'peaking', freq: 1500, gain: 3, Q: 3.0 },
        { type: 'highshelf', freq: 15000, gain: 5 },
         { type: 'highshelf', freq: 20000, gain: -6 },
    ];


     // ダイナミクス設定調整 - v8.0ベース維持
     const ULTIMATE_IMAX_COMPRESSOR_THRESHOLD_ADVANCED = -10;
    const ULTIMATE_IMAX_COMPRESSOR_RATIO_ADVANCED = 12;
    const ULTIMATE_IMAX_COMPRESSOR_ATTACK_ADVANCED = 0.0005;
    const ULTIMATE_IMAX_COMPRESSOR_RELEASE_ADVANCED = 1.0;
    const ULTIMATE_IMAX_COMPRESSOR_KNEE_ADVANCED = 20;

     const ULTIMATE_IMAX_EXPANDER_BOOST_GAIN_ADVANCED = 4.0;

     // ピークリミッター設定 - v8.0ベース維持
     const PEAK_LIMITER_THRESHOLD_DB_ADVANCED = 0;
     const PEAK_LIMITER_ATTACK_ADVANCED = 0;
     const PEAK_LIMITER_RELEASE_ADVANCED = 0.01;


    // 簡易ステレオワイド設定例 - v8.0ベース維持
    const STEREO_WIDE_DELAY_MS = 30;
    const STEREO_WIDE_LEFT_GAIN = 1.0;
    const STEREO_WIDE_RIGHT_GAIN = 0.85;

    // 簡易リバーブ設定例 (Convolutionと併用するため、補助的に調整) - v8.0ベース維持
    const REVERB_DELAY_SEC = 0.05;
    const REVERB_FEEDBACK = 0.4;
    const REVERB_WET_GAIN = 0.1;


    // dB値をゲイン値に変換するヘルパー関数
    function dBToGain(db) {
        return Math.pow(10, db / 20);
    }

    // --- 音響物理学応用モデル計算関数 ---
    // 評価計算部分は維持 (ログ出力なし)

    /**
     * Cross-sectional area of horn at distance x
     * S(x) = S0 * exp(m * x)
     */
    function S(x) {
        return HORN_S0 * Math.exp(HORN_M * x);
    }

    /**
     * Acoustic impedance at distance x
     * Z(x) = (rho0 * c) / S(x)
     */
    function Z(x) {
        return (RHO0 * C) / S(x);
    }

    /**
     * Exact plane wave pressure at (x, t)
     * p(x, t) = P0 * exp(j*(omegat - kx))
     * Returns a complex number as { re, im }
     */
    function pPlane(x, t, frequency, sourceAmplitude_P0) {
        const omega = 2 * Math.PI * frequency;
        const k = omega / C;
        const phase = omega * t - k * x;
        return {
            re: sourceAmplitude_P0 * Math.cos(phase),
            im: sourceAmplitude_P0 * Math.sin(phase)
        };
    }

    /**
     * Approximate pressure including area-expansion attenuation
     * p_approx(x, t) = P0 * sqrt(S0/S(x)) * exp(j*(omegat - kx))
     */
    function pApprox(x, t, frequency, sourceAmplitude_P0) {
        const factor = Math.sqrt(HORN_S0 / S(x));
        const omega = 2 * Math.PI * frequency;
        const k = omega / C;
        const phase = omega * t - k * x;
        return {
            re: sourceAmplitude_P0 * factor * Math.cos(phase),
            im: sourceAmplitude_P0 * factor * Math.sin(phase)
        };
    }


    // 音響管の共鳴特性計算関数 (簡易モデル)
    function calculateTubeResonanceGain(frequency, resonanceFrequency, speedOfSound, Q = 10) {
        const delta_f = Math.abs(frequency - resonanceFrequency);
        const bandwidth = resonanceFrequency / Q;
        return 1.0 / (1.0 + Math.pow(delta_f / (bandwidth / 2), 2));
    }

     // サブハーモニック合成ノードを作成する関数
    function createSubharmonicNode(audioCtx, ratio, gainValue) {
        const osc = audioCtx.createOscillator();
        osc.type = 'sine';
        const gain = audioCtx.createGain();
        gain.gain.value = gainValue;
        osc.connect(gain);
        return { osc, gain };
    }

     // 基本周波数を推定する関数 (自己相関法)
     function estimateFundamental(analyser) {
        if (!analyser) {
             console.warn("AnalyserNode not provided for fundamental frequency estimation.");
             return 0; // 推定失敗
        }
        const data = new Float32Array(analyser.fftSize);
        analyser.getFloatTimeDomainData(data);

        let bestOffset = -1;
        let bestCorr = -Infinity;
        const sampleRate = analyser.context.sampleRate;
        const maxOffset = Math.min(Math.floor(sampleRate / 20), data.length / 2); // 最低20Hzまで
        const minOffset = Math.max(2, Math.floor(sampleRate / 500)); // 最高500Hzまで

        for (let off = minOffset; off < maxOffset; off++) {
            let corr = 0;
            for (let i = 0; i < data.length - off; i++) {
                corr += data[i] * data[i + off];
            }
            if (corr > bestCorr) { bestCorr = corr; bestOffset = off; }
        }

        if (bestCorr > 0.1 * data.length && bestOffset > 0) { // 適当な閾値とオフセットチェック
            const fundamental = sampleRate / bestOffset;
             console.log(`Estimated fundamental frequency: ${fundamental.toFixed(2)} Hz (Offset: ${bestOffset})`);
            return fundamental;
        } else {
             console.log("Fundamental frequency estimation failed (correlation too low or offset zero).");
             return 0; // 推定失敗
        }
    }

    // 簡易合成IRデータを生成する関数 (新規)
     function createSyntheticIR(audioContext, duration, decayRate, numReflections, maxDelay) {
         const sampleRate = audioContext.sampleRate;
         const numSamples = Math.floor(duration * sampleRate);
         const buffer = audioContext.createBuffer(2, numSamples, sampleRate); // ステレオIR

         for (let channel = 0; channel < buffer.numberOfChannels; channel++) {
             const nowBuffering = buffer.getChannelData(channel);

             // 直接音 (最初のピーク)
             if (numSamples > 0) {
                  nowBuffering[0] = 1.0; // 最初のサンプルを1.0に

                  // 簡易的なランダム反射音を加える
                  for(let i = 0; i < numReflections; i++) {
                       const delaySamples = Math.floor(Math.random() * maxDelay * sampleRate / 1000); // ランダムなディレイ（サンプル数）
                       const gain = (Math.random() * 0.4 + 0.1) * Math.exp(-delaySamples / sampleRate * decayRate); // ランダムなゲインと減衰
                       if (delaySamples < numSamples) {
                           nowBuffering[delaySamples] += gain * (channel === 0 ? 1 : (Math.random() * 0.5 + 0.5)); // 左右チャンネルで少し異なるゲイン
                       }
                  }
             }
         }
         return buffer;
     }


    // --- AudioWorkletProcessor definition (Optimized LocalEnhance - Enhanced Interpolation) ---
    // LocalEnhance処理を高速化・リアルタイム向けに再設計（補間強化版）
    const processorCodeOptimized = `
    class LocalEnhanceProcessor extends AudioWorkletProcessor {
        // Worklet内の設定値（メインスレッドから渡すことも可能）
        LE_CHUNK_SIZE = ${LE_CHUNK_SIZE_WORKLET};
        LE_RMS_THRESHOLD = ${LE_RMS_THRESHOLD_WORKLET};
        EW_WAVELENGTH_FACTOR = ${EW_WAVELENGTH_FACTOR};
        EW_ENHANCEMENT_FACTOR = ${EW_ENHANCEMENT_FACTOR};
        GAMMA_L = ${GAMMA_L};
        GAMMA_T = ${GAMMA_T};
        GAMMA_MU0 = ${GAMMA_MU0};
        GAMMA_KAPPA = ${GAMMA_KAPPA};
        GAMMA_ALPHAP = ${GAMMA_ALPHAP};
        GAMMA_THRESHOLD = ${GAMMA_THRESHOLD_WORKLET};
        GAMMA_INTERPOLATION_FACTOR = ${GAMMA_INTERPOLATION_FACTOR};
        GAMMA_ENHANCE_GAIN = ${GAMMA_ENHANCE_GAIN};


        // チャンク処理と補間のためのバッファ（過去のサンプルを保持）
        // 補間には現在のサンプルとその前後のサンプルが必要
        chunkBuffer = []; // チャンネルごとのバッファ
        bufferLength = 128; // 標準の処理バッファ長 (Workletの入力/出力サイズ)
        historyBufferLength = this.LE_CHUNK_SIZE * 2; // RMS計算や補間に必要な過去のサンプル数（チャンクサイズの倍数程度）

        constructor() {
            super();
            // 初期化時にチャンクバッファを作成
            // Workletの入力バッファサイズ(128) + RMS計算や補間に必要な過去のサンプル数をバッファとして持つ
            this.chunkBuffer = new Array(this.numberOfInputs[0]);
            for(let i = 0; i < this.numberOfInputs[0]; i++) {
                 this.chunkBuffer[i] = new Float32Array(this.bufferLength + this.historyBufferLength);
                 // バッファをゼロで初期化
                 this.chunkBuffer[i].fill(0);
            }
             // console.log('[Worklet] Processor initialized.'); // Worklet内のログはブラウザコンソールとは別
        }

        // 簡易RMS計算
        calculateRMS(buffer) {
            let sum = 0;
            for (let i = 0; i < buffer.length; i++) {
                sum += buffer[i] * buffer[i];
            }
            return Math.sqrt(sum / buffer.length);
        }

        // 簡易3次補間 (toFixed(2) 廃止)
        cubicInterpolate(y0,y1,y2,y3,mu) {
            const a0=y3-y2-y0+y1;
            const a1=y0-y1-a0;
            const a2=y2-y0;
            const a3=y1;
            return a0*mu*mu*mu + a1*mu*mu + a2*mu + a3;
        }

        // Gamma処理の代替案（補間強化版）：特定の閾値を超える部分に補間と簡易ゲインを適用
        // サンプル挿入ではなく、既存サンプルの値を補間計算で調整
        applyGammaEnhancement(buffer, currentSampleIndexInBuffer) {
             const sample = buffer[currentSampleIndexInBuffer];
             const mu_i = Math.abs(sample);
             let gamma = 0;
             if (mu_i > 1e-9) {
                const logArg = Math.abs((Math.sqrt(this.GAMMA_MU0) - Math.sqrt(mu_i)) / (Math.sqrt(this.GAMMA_MU0) + Math.sqrt(mu_i)));
                if (logArg > 1e-9) {
                   gamma = -1 / this.GAMMA_L * Math.sqrt(this.GAMMA_T / mu_i) * Math.log(logArg) * (1 + this.GAMMA_KAPPA * this.GAMMA_ALPHAP / (this.GAMMA_L * this.GAMMA_L));
                }
             }
             const g = gamma; // toFixed(2) 廃止

             let processedSample = sample;

             if(Math.abs(g) > this.GAMMA_THRESHOLD){
               // Gamma条件を満たした場合、補間による波形調整を試みる
               // 現在のサンプルとその前後のサンプルを使用して補間値を計算
               // 簡易的な4点補間を応用
               const y0_idx = Math.max(0, currentSampleIndexInBuffer - 2);
               const y1_idx = Math.max(0, currentSampleIndexInBuffer - 1);
               const y2_idx = Math.min(buffer.length - 1, currentSampleIndexInBuffer + 1);
               const y3_idx = Math.min(buffer.length - 1, currentSampleIndexInBuffer + 2);

               const y0 = buffer[y0_idx];
               const y1 = buffer[y1_idx];
               const y2 = buffer[y2_idx];
               const y3 = buffer[y3_idx];

               // 現在のサンプル位置における補間値を計算 (mu=1.0 で現在の位置の補間値を直接計算)
               // または、現在のサンプルと次のサンプルの間 (mu=0.5) の補間値を計算し、それを混ぜるなど
               // ここでは、現在のサンプルと次のサンプル間の補間を計算し、元のサンプルとブレンドする簡易的な方法
               const mu = 0.5; // 現在のサンプルと次のサンプルの中間
               const interpolatedValue = this.cubicInterpolate(y0, y1, sample, y2, mu); // y2, y3の代わりにsample, y2を使用 (位置関係を調整)

               // 元のサンプルと補間値をブレンド
               processedSample = sample * (1.0 - this.GAMMA_INTERPOLATION_FACTOR) + interpolatedValue * this.GAMMA_INTERPOLATION_FACTOR;

               // 補間適用時の追加ゲインを適用
               processedSample *= this.GAMMA_ENHANCE_GAIN;
             }
             return processedSample;
        }

        process(inputs, outputs) {
            const input = inputs[0];
            const output = outputs[0];
            const channelCount = input ? input.length : 0;

            if (channelCount > 0) {
                for(let ch = 0; ch < channelCount; ch++) {
                    const inputChannel = input[ch];
                    const outputChannel = output[ch];

                    if (inputChannel && outputChannel) {
                        const inputLength = inputChannel.length; // 通常128

                        // チャンクバッファの更新（新しいデータをバッファに追加し、古いデータを捨てる）
                        // 新しい入力データをバッファの最後に追加
                        const buffer = this.chunkBuffer[ch];
                        buffer.set(buffer.subarray(inputLength), 0); // 古いデータを前に詰める
                        buffer.set(inputChannel, this.historyBufferLength); // 新しいデータをhistoryBufferLength以降に追加 (これで入力バッファの位置が固定される)


                        // Workletの処理ブロックごとにLocalEnhance処理を適用
                        for (let i = 0; i < inputLength; i++) {
                            // バッファ全体における現在の入力サンプルの位置
                            const currentSampleIndexInBuffer = this.historyBufferLength + i;

                            // ローカルエンハンスメント処理の簡略化・再設計
                            // RMS判定はチャンク単位で行うが、適用はサンプルごと
                            // RMS計算用のチャンクは、現在のサンプルを中心とする範囲をバッファから切り出す
                             const rmsChunkStartIndex = Math.max(0, currentSampleIndexInBuffer - Math.floor(this.LE_CHUNK_SIZE / 2));
                             const rmsChunkEndIndex = Math.min(buffer.length, rmsChunkStartIndex + this.LE_CHUNK_SIZE);
                             const rmsChunk = buffer.subarray(rmsChunkStartIndex, rmsChunkEndIndex);

                            const rms = this.calculateRMS(rmsChunk);

                            let processedSample = inputChannel[i]; // デフォルトは元のサンプル

                            if (rms < this.LE_RMS_THRESHOLD) {
                                // 弾性波動模倣とGamma処理（補間強化版）を適用

                                // 簡易弾性波動模倣 (リアルタイム対応)
                                // Workletバッファ全体ではなく、現在の入力バッファ内での位置に基づく簡易w計算
                                const w = (i / inputLength) * this.EW_ENHANCEMENT_FACTOR;
                                const targetValue = buffer[currentSampleIndexInBuffer]; // バッファ全体から現在のサンプル値を取得
                                const preValue = buffer[Math.max(0, currentSampleIndexInBuffer - 1)]; // バッファ全体から前のサンプル値を取得
                                processedSample = preValue * (1 - w) + targetValue * w;


                                // Gamma処理代替（補間強化版）を適用
                                processedSample = this.applyGammaEnhancement(buffer, currentSampleIndexInBuffer); // バッファ全体と現在のインデックスを渡す
                                // applyGammaEnhancement関数内で、rmsや他の条件も考慮して処理するか、
                                // あるいは外側でrms条件を満たす場合のみ applyGammaEnhancement を呼ぶか

                                // v7.0のロジックを踏襲し、RMS閾値以下の場合に弾性波動とGamma代替を適用
                                const preIndex_ew = Math.max(0, currentSampleIndexInBuffer - 1);
                                const targetIndex_ew = currentSampleIndexInBuffer;
                                const preValue_ew = buffer[preIndex_ew];
                                const targetValue_ew = buffer[targetIndex_ew];
                                const w_ew = (i / inputLength) * this.EW_ENHANCEMENT_FACTOR; // Workletバッファ内の相対位置 i を使用 (0-127)
                                let sampleAfterEW = preValue_ew * (1 - w_ew) + targetValue_ew * w_ew;

                                // Gamma条件チェックと補間強化処理の適用
                                const sampleAfterGamma = this.applyGammaEnhancement(buffer, currentSampleIndexInBuffer);

                                // ここで、弾性波動の結果とGamma代替の結果をどのように組み合わせるか決定
                                // 例えば、Gamma条件を満たす場合はGamma代替の結果を優先するなど
                                // シンプルに、RMS条件を満たす場合に、弾性波動とGamma代替の両方を適用した結果とする
                                // あるいは、Gamma条件を満たす場合はapplyGammaEnhancementの結果を、満たさない場合は弾性波動の結果を使用するなど
                                // ここでは、RMS条件を満たす場合に、applyGammaEnhancementの結果（内部で補間と簡易ゲイン適用）を使用する
                                processedSample = sampleAfterGamma; // applyGammaEnhancement 内で RMS閾値以下の条件を考慮
                                // applyGammaEnhancement 関数自体を修正し、RMS条件を内部でチェックするように変更が必要

                                // applyGammaEnhancement 関数の引数とロジックを変更
                                // function applyGammaEnhancement(buffer, currentSampleIndexInBuffer, rms) { ... }
                                // if (rms < this.LE_RMS_THRESHOLD && Math.abs(g) > this.GAMMA_THRESHOLD) { ... }

                                // 上記の再考を踏まえ、processメソッド内のロジックを修正
                                // RMS条件を満たすかどうかで分岐
                                if (rms < this.LE_RMS_THRESHOLD) {
                                    // RMS条件を満たす場合
                                    // 弾性波動模倣を適用
                                    const preIndex_ew = Math.max(0, currentSampleIndexInBuffer - 1);
                                    const targetIndex_ew = currentSampleIndexInBuffer;
                                    const preValue_ew = buffer[preIndex_ew];
                                    const targetValue_ew = buffer[targetIndex_ew];
                                     // Workletバッファ内の相対位置 i を使用 (0-127) を0-1に正規化してw計算
                                    const w_ew = (i / inputLength) * this.EW_ENHANCEMENT_FACTOR;
                                    let sampleAfterEW = preValue_ew * (1 - w_ew) + targetValue_ew * w_ew;


                                    // Gamma条件を満たすかどうかをチェックし、満たす場合は補間強化処理を適用
                                     const mu_i = Math.abs(sampleAfterEW); // 弾性波動適用後のサンプル値でGamma条件をチェック
                                     let gamma = 0;
                                     if (mu_i > 1e-9) {
                                         const logArg = Math.abs((Math.sqrt(this.GAMMA_MU0) - Math.sqrt(mu_i)) / (Math.sqrt(this.GAMMA_MU0) + Math.sqrt(mu_i)));
                                         if (logArg > 1e-9) {
                                             gamma = -1 / this.GAMMA_L * Math.sqrt(this.GAMMA_T / mu_i) * Math.log(logArg) * (1 + this.GAMMA_KAPPA * this.GAMMA_ALPHAP / (this.GAMMA_L * this.GAMMA_L));
                                         }
                                     }
                                     const g = gamma;

                                    if (Math.abs(g) > this.GAMMA_THRESHOLD) {
                                         // Gamma条件を満たす場合、補間による波形調整を適用
                                         // 現在のサンプルとその前後のサンプルを使用して補間値を計算
                                         const y0_idx = Math.max(0, currentSampleIndexInBuffer - 2);
                                         const y1_idx = Math.max(0, currentSampleIndexInBuffer - 1);
                                         const y2_idx = Math.min(buffer.length - 1, currentSampleIndexInBuffer + 1);
                                         const y3_idx = Math.min(buffer.length - 1, currentSampleIndexInBuffer + 2);

                                         const y0 = buffer[y0_idx];
                                         const y1 = buffer[y1_idx];
                                         const y2 = buffer[y2_idx];
                                         const y3 = buffer[y3_idx];

                                         // 現在のサンプルと次のサンプルの中間 (mu=0.5) の補間値を計算し、弾性波動適用後のサンプルとブレンド
                                         const mu = 0.5;
                                         // 補間関数に渡す y0, y1, y2, y3 を、バッファ全体から取得した値に修正
                                         const interpolatedValue = this.cubicInterpolate(y0, y1, targetValue_ew, y2, mu); // 弾性波動前の元の値を補間に使用

                                         // 弾性波動適用後のサンプルと補間値をブレンド
                                         processedSample = sampleAfterEW * (1.0 - this.GAMMA_INTERPOLATION_FACTOR) + interpolatedValue * this.GAMMA_INTERPOLATION_FACTOR;

                                         // 補間適用時の追加ゲインを適用
                                         processedSample *= this.GAMMA_ENHANCE_GAIN;

                                    } else {
                                         // Gamma条件を満たさない場合は、弾性波動適用後のサンプルをそのまま使用
                                         processedSample = sampleAfterEW;
                                    }

                                } else {
                                    // RMS条件を満たさない場合は、元の入力サンプルをそのまま使用
                                    processedSample = inputChannel[i];
                                }

                            } else {
                                // RMS条件を満たさない場合は、元の入力サンプルをそのまま使用
                                processedSample = inputChannel[i];
                            }

                            outputChannel[i] = processedSample;
                        }

                    } else {
                         if (outputChannel) outputChannel.fill(0);
                    }
                }
            }
            return true;
        }
    }
    registerProcessor('local-enhance-processor', LocalEnhanceProcessor);
    `;

    const blobOptimized = new Blob([processorCodeOptimized], { type: 'application/javascript' });
    let moduleURLOptimized = URL.createObjectURL(blobOptimized);

    const activeAudioContexts = [];
    let realTimeAudioContext = null;
    let mediaSourceNode = null;
    let workletNode = null;


     // リアルタイムAudioContextをセットアップし、ノードグラフを構築・接続
    async function setupRealtimeAudio(media, fundamentalFrequency, physicalEvaluationResults) {
        console.log("Setting up realtime AudioContext and node graph...");

        if (realTimeAudioContext && realTimeAudioContext.state !== 'closed') {
            await realTimeAudioContext.close().catch(e => console.error("Failed to close existing AudioContext:", e));
            activeAudioContexts.splice(activeAudioContexts.indexOf(realTimeAudioContext), 1);
            realTimeAudioContext = null;
        }

        realTimeAudioContext = new (window.AudioContext || window.webkitAudioContext)({
             sampleRate: TARGET_SAMPLE_RATE
        });
        activeAudioContexts.push(realTimeAudioContext);
         console.log(`Created Realtime AudioContext with sample rate: ${realTimeAudioContext.sampleRate} Hz (Target: ${TARGET_SAMPLE_RATE} Hz).`);


        try {
            await realTimeAudioContext.audioWorklet.addModule(moduleURLOptimized);
            console.log("Optimized AudioWorklet module (Enhanced Gamma) added.");
        } catch (e) {
            console.error("Failed to add Optimized AudioWorklet module:", e);
            throw new Error("Failed to add AudioWorklet module.");
        }

        mediaSourceNode = realTimeAudioContext.createMediaElementSource(media);
         console.log(`MediaElementSourceNode created. Channels: ${mediaSourceNode.channelCount}`);

        workletNode = new AudioWorkletNode(realTimeAudioContext, 'local-enhance-processor', {
            numberOfInputs: 1,
            numberOfOutputs: 1,
            outputChannelCount: [mediaSourceNode.channelCount]
        });
         console.log("AudioWorkletNode (local-enhance-processor) created.");


        // --- リアルタイムAudioContext ノードグラフ構築 ---

        // メディアソース -> Worklet (LocalEnhance)
        let currentNode = mediaSourceNode;
         currentNode.connect(workletNode);
         currentNode = workletNode;

         if (mediaSourceNode.channelCount > 0) {

             // Worklet出力 -> 簡易ステレオワイド -> Mid/Side分解 -> ...

             // 簡易ステレオワイド処理ノード
             const splitterSimStereo_rt = realTimeAudioContext.createChannelSplitter(mediaSourceNode.channelCount);
             const mergerSimStereo_rt = realTimeAudioContext.createChannelMerger(mediaSourceNode.channelCount);
             const delayRightSimStereo_rt = realTimeAudioContext.createDelay(1.0);
             delayRightSimStereo_rt.delayTime.setValueAtTime(STEREO_WIDE_DELAY_MS / 1000, realTimeAudioContext.currentTime);
             const gainLeftSimStereo_rt = realTimeAudioContext.createGain();
             gainLeftSimStereo_rt.gain.setValueAtTime(STEREO_WIDE_LEFT_GAIN, realTimeAudioContext.currentTime);
             const gainRightSimStereo_rt = realTimeAudioContext.createGain();
             gainRightSimStereo_rt.gain.setValueAtTime(STEREO_WIDE_RIGHT_GAIN, realTimeAudioContext.currentTime);

              currentNode.connect(splitterSimStereo_rt);
              if (splitterSimStereo_rt.numberOfOutputs > 0) {
                   splitterSimStereo_rt.connect(gainLeftSimStereo_rt, 0);
                   gainLeftSimStereo_rt.connect(mergerSimStereo_rt, 0, 0);
                   if (splitterSimStereo_rt.numberOfOutputs > 1) {
                      splitterSimStereo_rt.connect(delayRightSimStereo_rt, 1);
                      delayRightSimStereo_rt.connect(gainRightSimStereo_rt);
                      gainRightSimStereo_rt.connect(mergerSimStereo_rt, 0, 1);
                   } else {
                       splitterSimStereo_rt.connect(delayRightSimStereo_rt, 0);
                       delayRightSimStereo_rt.connect(gainRightSimStereo_rt);
                       gainRightSimStereo_rt.connect(mergerSimStereo_rt, 0, 1);
                   }
              }
             currentNode = mergerSimStereo_rt;


             // Mid/Side 処理ノード群
             const splitterMS_rt = realTimeAudioContext.createChannelSplitter(mediaSourceNode.channelCount);
             const mergerMS_rt = realTimeAudioContext.createChannelMerger(mediaSourceNode.channelCount);
             const midGain_rt = realTimeAudioContext.createGain(); midGain_rt.gain.setValueAtTime(0.5, realTimeAudioContext.currentTime);
             const sideGain_rt = realTimeAudioContext.createGain(); sideGain_rt.gain.setValueAtTime(0.5, realTimeAudioContext.currentTime);
             const negSideGain_rt = realTimeAudioContext.createGain(); negSideGain_rt.gain.setValueAtTime(-0.5, realTimeAudioContext.currentTime);
             const midProcessGain_rt = realTimeAudioContext.createGain();
             midProcessGain_rt.gain.setValueAtTime(dBToGain(MID_GAIN_DB), realTimeAudioContext.currentTime);
             const sideProcessGain_rt = realTimeAudioContext.createGain();
             sideProcessGain_rt.gain.setValueAtTime(dBToGain(SIDE_GAIN_DB), realTimeAudioContext.currentTime);
             const gainNegOne_rt = realTimeAudioContext.createGain(); gainNegOne_rt.gain.setValueAtTime(-1.0, realTimeAudioContext.currentTime);

             currentNode.connect(splitterMS_rt);
             splitterMS_rt.connect(midGain_rt, 0);
             if (mediaSourceNode.channelCount > 1) splitterMS_rt.connect(midGain_rt, 1);
             splitterMS_rt.connect(sideGain_rt, 0);
             if (mediaSourceNode.channelCount > 1) {
                 splitterMS_rt.connect(negSideGain_rt, 1);
                 negSideGain_rt.connect(sideGain_rt);
             }
             midGain_rt.connect(midProcessGain_rt);
             sideGain_rt.connect(sideProcessGain_rt);
             midProcessGain_rt.connect(mergerMS_rt, 0, 0);
             sideProcessGain_rt.connect(mergerMS_rt, 0, 0);
             if (mediaSourceNode.channelCount > 1) {
                  midProcessGain_rt.connect(mergerMS_rt, 0, 1);
                  sideProcessGain_rt.connect(gainNegOne_rt, 0);
                  gainNegOne_rt.connect(mergerMS_rt, 0, 1);
             }
             currentNode = mergerMS_rt;


             // Mid/Side出力 -> EQ補正群 -> 20Hz低音EQ -> 簡易エキスパンダー風Gain -> 物理応用フィルター群

             // EQ補正群 (IMAX向け調整) - ULTIMATE_IMAX_EQ_SETTINGS_ADVANCED を使用
             const imaxEQs_rt = ULTIMATE_IMAX_EQ_SETTINGS_ADVANCED.map(settings => {
                  const eq = realTimeAudioContext.createBiquadFilter();
                  eq.type = settings.type;
                  eq.frequency.setValueAtTime(settings.freq, realTimeAudioContext.currentTime);
                  eq.gain.setValueAtTime(settings.gain, realTimeAudioContext.currentTime);
                  if (settings.Q) eq.Q.setValueAtTime(settings.Q, realTimeAudioContext.currentTime);
                  return eq;
              });

             imaxEQs_rt.forEach(eq => {
                  currentNode.connect(eq);
                  currentNode = eq;
              });

             // 20Hzまでの低音再現 (低域強調EQ) - 設定値は ULTIMATE... から流用
             const eqLow_rt = realTimeAudioContext.createBiquadFilter();
             eqLow_rt.type = 'lowshelf';
             eqLow_rt.frequency.setValueAtTime(ULTIMATE_IMAX_EQ_SETTINGS_ADVANCED[0].freq, realTimeAudioContext.currentTime);
             eqLow_rt.gain.setValueAtTime(ULTIMATE_IMAX_EQ_SETTINGS_ADVANCED[0].gain, realTimeAudioContext.currentTime);
             currentNode.connect(eqLow_rt);
             currentNode = eqLow_rt;


             // 簡易エキスパンダー風ゲインノード - ULTIMATE_IMAX_EXPANDER_BOOST_GAIN_ADVANCED を使用
             const expanderGain_rt = realTimeAudioContext.createGain();
             expanderGain_rt.gain.setValueAtTime(ULTIMATE_IMAX_EXPANDER_BOOST_GAIN_ADVANCED, realTimeAudioContext.currentTime);
             currentNode.connect(expanderGain_rt);
             currentNode = expanderGain_rt;


             // 物理応用フィルター群を接続
             const hornCutoffFilter_rt = realTimeAudioContext.createBiquadFilter();
             hornCutoffFilter_rt.type = 'highpass';
             hornCutoffFilter_rt.frequency.setValueAtTime(HORN_CUTOFF_FREQ_CALCULATED, realTimeAudioContext.currentTime);
             hornCutoffFilter_rt.Q.setValueAtTime(0.5, realTimeAudioContext.currentTime);

             const tubeResonanceCompFilter_rt = realTimeAudioContext.createBiquadFilter();
             tubeResonanceCompFilter_rt.type = 'notch';
             tubeResonanceCompFilter_rt.frequency.setValueAtTime(TUBE_RESONANCE_FREQ, realTimeAudioContext.currentTime);
             tubeResonanceCompFilter_rt.Q.setValueAtTime(10.0, realTimeAudioContext.currentTime);

             currentNode.connect(hornCutoffFilter_rt);
             hornCutoffFilter_rt.connect(tubeResonanceCompFilter_rt);
             currentNode = tubeResonanceCompFilter_rt;


             // 簡易空間反射群 (Convolutionと併用)
             const earlyReflectionDelays_rt = EARLY_REFLECTION_DELAYS_MS.map(delayMs => {
                  const delay = realTimeAudioContext.createDelay(1.0);
                  delay.delayTime.setValueAtTime(delayMs / 1000, realTimeAudioContext.currentTime);
                  return delay;
              });
             const earlyReflectionGains_rt = EARLY_REFLECTION_GAINS.map(gainValue => {
                  const gain = realTimeAudioContext.createGain();
                  gain.gain.setValueAtTime(gainValue, realTimeAudioContext.currentTime);
                  return gain;
              });
              const earlyReflectionMerger_rt = realTimeAudioContext.createChannelMerger(mediaSourceNode.channelCount);

             const dryGain_rt = realTimeAudioContext.createGain();
             dryGain_rt.gain.setValueAtTime(CONVOLVER_DRY_GAIN, realTimeAudioContext.currentTime); // Convolutionのドライゲインと兼ねる

             // ConvolutionNode (新規追加)
             const convolverNode = realTimeAudioContext.createConvolver();
              // 合成IRデータを生成してBufferにセット
              try {
                  convolverNode.buffer = createSyntheticIR(
                      realTimeAudioContext,
                      CONVOLVER_IR_DURATION_SEC,
                      CONVOLVER_IR_DECAY_RATE,
                      CONVOLVER_IR_NUM_REFLECTIONS,
                      EARLY_REFLECTION_DELAYS_MS[EARLY_REFLECTION_DELAYS_MS.length - 1] || 50
                  );
                   console.log("Synthetic Impulse Response created and loaded into ConvolverNode.");
              } catch (e) {
                  console.error("Failed to create or load synthetic IR for ConvolverNode:", e);
              }

             const convolverWetGain = realTimeAudioContext.createGain();
             convolverWetGain.gain.setValueAtTime(CONVOLVER_WET_GAIN, realTimeAudioContext.currentTime);


             // 簡易リバーブ処理 (フィードバックディレイ) - Convolutionと併用
             const reverbDelay_rt = realTimeAudioContext.createDelay(1.0);
             reverbDelay_rt.delayTime.setValueAtTime(REVERB_DELAY_SEC, realTimeAudioContext.currentTime);
             const reverbFeedbackGain_rt = realTimeAudioContext.createGain();
             reverbFeedbackGain_rt.gain.setValueAtTime(REVERB_FEEDBACK, realTimeAudioContext.currentTime);
             const reverbWetGain_rt = realTimeAudioContext.createGain();
             reverbWetGain_rt.gain.setValueAtTime(REVERB_WET_GAIN, realTimeAudioContext.currentTime);
             reverbDelay_rt.connect(reverbFeedbackGain_rt);
             reverbFeedbackGain_rt.connect(reverbDelay_rt);


             // 各成分を最終ミックスMergerに接続
             const finalMixMerger = realTimeAudioContext.createChannelMerger(mediaSourceNode.channelCount);


             // 物理応用フィルター出力 ->
             // ドライ信号（Convolutionのドライゲイン）
             currentNode.connect(dryGain_rt);
             dryGain_rt.connect(finalMixMerger, 0, 0);
             dryGain_rt.connect(finalMixMerger, 0, 1); // ステレオ


             // -> 簡易反射ノード群 (ミックス後、最終ミックスMergerへ)
             let currentNodeForReflections = currentNode; // 物理応用フィルターの出力から分岐
             earlyReflectionDelays_rt.forEach((delayNode, index) => {
                  if (index < earlyReflectionGains_rt.length) {
                      currentNodeForReflections.connect(delayNode);
                      delayNode.connect(earlyReflectionGains_rt[index]);
                      earlyReflectionGains_rt[index].connect(earlyReflectionMerger_rt, 0, 0);
                      earlyReflectionGains_rt[index].connect(earlyReflectionMerger_rt, 0, 1);
                  }
              });
             earlyReflectionMerger_rt.connect(finalMixMerger, 0, 0);
             earlyReflectionMerger_rt.connect(finalMixMerger, 0, 1);


             // -> Convolution Node (ウェット信号)
             let currentNodeForConvolver = currentNode; // 物理応用フィルターの出力から分岐
             if (convolverNode.buffer) { // IRがロードできていれば接続
                 currentNodeForConvolver.connect(convolverNode);
                 convolverNode.connect(convolverWetGain);
                 convolverWetGain.connect(finalMixMerger, 0, 0);
                 convolverWetGain.connect(finalMixMerger, 0, 1);
             } else {
                 console.warn("ConvolutionNode buffer is not set. Skipping Convolution.");
             }


             // -> 簡易リバーブ (ウェット信号)
             let currentNodeForReverb = currentNode; // 物理応用フィルターの出力から分岐
             currentNodeForReverb.connect(reverbDelay_rt);
             reverbDelay_rt.connect(reverbWetGain_rt);
             reverbWetGain_rt.connect(finalMixMerger, 0, 0);
             reverbWetGain_rt.connect(finalMixMerger, 0, 1);


             currentNode = finalMixMerger; // ミックスされた信号が次へ


             // ミックス出力 -> 全体のダイナミクス調整 -> ピークリミッター -> ラウドネス調整Gain -> Destination

             // 全体のダイナミクス調整用コンプレッサー
             const compressor_rt = realTimeAudioContext.createDynamicsCompressor();
             compressor_rt.threshold.setValueAtTime(ULTIMATE_IMAX_COMPRESSOR_THRESHOLD_ADVANCED, realTimeAudioContext.currentTime);
             compressor_rt.ratio.setValueAtTime(ULTIMATE_IMAX_COMPRESSOR_RATIO_ADVANCED, realTimeAudioContext.currentTime);
             compressor_rt.attack.setValueAtTime(ULTIMATE_IMAX_COMPRESSOR_ATTACK_ADVANCED, realTimeAudioContext.currentTime);
             compressor_rt.release.setValueAtTime(ULTIMATE_IMAX_COMPRESSOR_RELEASE_ADVANCED, realTimeAudioContext.currentTime);
             compressor_rt.knee.setValueAtTime(ULTIMATE_IMAX_COMPRESSOR_KNEE_ADVANCED, realTimeAudioContext.currentTime);
             currentNode.connect(compressor_rt);
             currentNode = compressor_rt;


             // サブハーモニック信号をコンプレッサーの入力に加算
             let subOsc_rt = null; // オシレーターをこのスコープで宣言
              if (fundamentalFrequency > 0) {
                  const { osc, gain } = createSubharmonicNode(realTimeAudioContext, SUBHARMONIC_RATIO, SUBHARMONIC_GAIN);
                  subOsc_rt = osc; // 参照を保持
                  subOsc_rt.frequency.setValueAtTime(fundamentalFrequency / SUBHARMONIC_RATIO, realTimeAudioContext.currentTime);
                  // サブハーモニック信号をコンプレッサーの入力に接続
                  gain.connect(compressor_rt);
                   console.log("Subharmonic signal connected to compressor input.");
                   // リアルタイムコンテキストの状態変化に合わせてオシレーターを開始
                   if (realTimeAudioContext.state === 'running') {
                       subOsc_rt.start(); // Contextがrunningなら即開始
                   } else {
                       // context.resume() 時に開始する必要がある
                   }
              } else {
                   console.warn("Fundamental frequency not estimated or is zero. Subharmonic oscillator will not be used.");
              }


             // ピークリミッター
             const peakLimiter_rt = realTimeAudioContext.createDynamicsCompressor();
             peakLimiter_rt.threshold.setValueAtTime(PEAK_LIMITER_THRESHOLD_DB_ADVANCED, realTimeAudioContext.currentTime);
             peakLimiter_rt.ratio.setValueAtTime(100, realTimeAudioContext.currentTime);
             peakLimiter_rt.attack.setValueAtTime(PEAK_LIMITER_ATTACK_ADVANCED, realTimeAudioContext.currentTime);
             peakLimiter_rt.release.setValueAtTime(PEAK_LIMITER_RELEASE_ADVANCED, realTimeAudioContext.currentTime);
             peakLimiter_rt.knee.setValueAtTime(0, realTimeAudioContext.currentTime);
             currentNode.connect(peakLimiter_rt);
             currentNode = peakLimiter_rt;


             // ラウドネス調整用ゲイン
             const finalGainNode_rt = realTimeAudioContext.createGain();
             finalGainNode_rt.gain.setValueAtTime(1.0, realTimeAudioContext.currentTime);
             currentNode.connect(finalGainNode_rt);
             currentNode = finalGainNode_rt;


             // 最終出力をDestinationに接続
             currentNode.connect(realTimeAudioContext.destination);

              console.log("Realtime AudioContext node graph connected to Destination.");

         } else {
             console.warn("Media source has zero channels. Connecting source directly to destination (no processing).");
             mediaSourceNode.connect(realTimeAudioContext.destination);
         }


        // AudioContextの状態がsuspendedの場合はresumeを試みる
         if (realTimeAudioContext.state === 'suspended') {
             console.log("AudioContext is suspended. Attempting to resume...");
             try {
                await realTimeAudioContext.resume();
                console.log("AudioContext resumed.");
                // Contextがrunningになったらサブハーモニックオシレーターを開始
                 // subOsc_rt 参照は setupRealtimeAudio スコープ内にあるため、ここで開始
                 if (fundamentalFrequency > 0 && subOsc_rt && subOsc_rt.state === 'pending') { // pending状態なら開始
                     subOsc_rt.start();
                      console.log("Subharmonic oscillator started after context resumed.");
                 }
             } catch (e) {
                 console.error("Failed to resume AudioContext (user interaction may be required):", e);
             }
         } else if (realTimeAudioContext.state === 'running') {
             // Contextが既にrunningの場合はここでオシレーターを開始
              if (fundamentalFrequency > 0 && subOsc_rt && subOsc_rt.state === 'pending') { // pending状態なら開始
                   subOsc_rt.start();
                    console.log("Subharmonic oscillator started (context was already running).");
              }
         }

         console.log("Realtime audio setup complete.");
         // setupRealtimeAudio 完了後、ユーザーがメディア要素を再生/一時停止することで処理済みの音が出る
    }


    // AudioContextと関連リソースのクリーンアップ
    async function cleanupRealtimeAudio() {
        console.log("Cleaning up realtime AudioContext resources...");
        // サブハーモニックオシレーターの停止（setupRealtimeAudioスコープ外からの参照が必要）
        // ここでは、リアルタイムAudioContextをクローズすることで、それに接続されている全てのノード（オシレーター含む）が停止/解放されることを期待する

        // ノード接続を解除
        if (mediaSourceNode) { mediaSourceNode.disconnect(); mediaSourceNode = null; }
        if (workletNode) { workletNode.disconnect(); workletNode = null; }
        // その他のリアルタイムノードも、数が多い場合は個別の参照保持と切断は現実的ではない
        // Contextクローズ時に自動的に解放されるのが理想的

        // AudioContextをクローズ
        if (realTimeAudioContext && realTimeAudioContext.state !== 'closed') {
            await realTimeAudioContext.close().catch(e => console.error("Failed to close AudioContext during cleanup:", e));
             console.log("Realtime AudioContext closed.");
        }
        realTimeAudioContext = null;

        // Blob URLを解放
        if (moduleURLOptimized) {
            URL.revokeObjectURL(moduleURLOptimized);
            moduleURLOptimized = null;
             console.log("AudioWorklet Blob URL revoked.");
        }
         console.log("Realtime audio cleanup complete.");
    }


    // --- リソース解放処理 ---
    window.addEventListener('unload', async () => {
        console.log("Page unloading. Performing cleanup.");
         // リアルタイムAudioContextのリソースを解放
        await cleanupRealtimeAudio();

        // アクティブな全てのAudioContextをクローズ (念のため、他のコンテキストもあれば)
        activeAudioContexts.forEach(ctx => {
            if (ctx && ctx.state !== 'closed') {
                ctx.close().catch(() => {});
            }
        });
        activeAudioContexts.length = 0;
         console.log("Unload cleanup complete.");
    });

     // エラー発生時のクリーンアップ
     const cleanupOnError = async (e) => {
         console.error("An error occurred during execution:", e);
         // リアルタイムAudioContextのリソースを解放
         await cleanupRealtimeAudio();

         // アクティブな全てのAudioContextをクローズ
         activeAudioContexts.forEach(ctx => {
             if (ctx && ctx.state !== 'closed') {
                 ctx.close().catch(() => {});
             }
         });
         activeAudioContexts.length = 0;
         console.log("Error cleanup complete.");
     };


    // UserScript実行開始
    try {
        const media = await waitForMedia();
        await primeDecode(media, PRIME_DECODE_WAIT_SEC); // Autoplay policy対策

        // 物理モデルの評価計算 (一度実行)
         const physicalEvaluationResults = {};
         // 評価例: 特定周波数 (例: 1000 Hz) におけるホーン出口 (x=L) での音圧計算
         const evaluationFrequency = 1000;
         const sampleCountForEvaluation = 10;
         const sourceAmplitude_P0 = 1.0;
         physicalEvaluationResults.hornExitPressure = [];
         for (let i = 0; i <= sampleCountForEvaluation; i++) {
            const t = (i / sampleCountForEvaluation) * (1.0 / evaluationFrequency) * 0.1 ;
            const exact = pPlane(HORN_L, t, evaluationFrequency, sourceAmplitude_P0);
            const approx = pApprox(HORN_L, t, evaluationFrequency, sourceAmplitude_P0);
            physicalEvaluationResults.hornExitPressure.push({ t: t, exact: exact, approx: approx });
         }
         const evalDistance_x = HORN_L / 2;
         physicalEvaluationResults.impedanceAtMidHorn = Z(evalDistance_x);
         const evalFreqTube1 = TUBE_RESONANCE_FREQ * 0.8;
         const evalFreqTube2 = TUBE_RESONANCE_FREQ;
         const evalFreqTube3 = TUBE_RESONANCE_FREQ * 1.2;
         physicalEvaluationResults.tubeResonanceGain = {
             freq1: { freq: evalFreqTube1, gain: calculateTubeResonanceGain(evalFreqTube1, TUBE_RESONANCE_FREQ, C) },
             freq2: { freq: evalFreqTube2, gain: calculateTubeResonanceGain(evalFreqTube2, TUBE_RESONANCE_FREQ, C) },
             freq3: { freq: evalFreqTube3, gain: calculateTubeResonanceGain(evalFreqTube3, TUBE_RESONANCE_FREQ, C) }
         };
         console.log("Physical model evaluation completed (results stored in variable).");


        // 基本周波数推定のためのデータ収集と推定
        console.log("Starting f0 estimation process...");
        const tempCtxForF0 = new AudioContext();
        activeAudioContexts.push(tempCtxForF0);
        const tempSourceForF0 = tempCtxForF0.createMediaElementSource(media);
        const analyserNodeForF0 = tempCtxForF0.createAnalyser();
        analyserNodeForF0.fftSize = 2048;

        tempSourceForF0.connect(analyserNodeForF0);

         const wasPausedForF0 = media.paused;
         let playAttemptSuccessfulForF0 = false;
         if (wasPausedForF0) {
             try {
                 const playPromise = media.play();
                 if (playPromise !== undefined) {
                     await playPromise;
                 }
                 playAttemptSuccessfulForF0 = true;
                 console.log("Media play started for f0 estimation data tapping.");
             } catch (e) {
                  console.warn("Media play failed for f0 estimation (may require user interaction):", e);
             }
         } else {
               playAttemptSuccessfulForF0 = true;
               console.log("Media was already playing. Starting f0 estimation data tapping.");
         }

        await new Promise(r => setTimeout(r, TAP_AUDIO_DURATION_SEC * 1000));
        console.log(`Finished f0 estimation data tapping (${TAP_AUDIO_DURATION_SEC} seconds).`);

        const fundamentalFrequency = estimateFundamental(analyserNodeForF0);
        console.log(`Estimated f0: ${fundamentalFrequency.toFixed(2)} Hz`);

        tempSourceForF0.disconnect();
        analyserNodeForF0.disconnect();
        if (tempCtxForF0 && tempCtxForF0.state !== 'closed') {
             tempCtxForF0.close().catch(e => console.error("Failed to close temporary AudioContext for f0:", e));
        }
        activeAudioContexts.splice(activeAudioContexts.indexOf(tempCtxForF0), 1);
         console.log("Temporary AudioContext for f0 estimation closed.");

        if (wasPausedForF0 && playAttemptSuccessfulForF0) {
            media.pause();
            console.log("Media paused after f0 estimation data tapping.");
        } else if (!wasPausedForF0 && playAttemptSuccessfulForF0) {
             console.log("Media continued playing after f0 estimation data tapping.");
        } else if (!playAttemptSuccessfulForF0) {
              console.warn("Media play for f0 estimation failed, original state might be affected.");
        }


        // リアルタイムAudioContextをセットアップし、ノードグラフを構築・接続
        await setupRealtimeAudio(media, fundamentalFrequency, physicalEvaluationResults);

        console.log("Audio processing graph is ready. Control playback using the media element.");
        console.log("If no sound is heard, try interacting with the page (e.g., click, play media) to resume the AudioContext.");


    } catch (e) {
        // エラー発生時はクリーンアップ
        cleanupOnError(e);
    }

})();

