// ==UserScript==
// @name         Equalizer Nomal
// @namespace    maid.aural.inspection
// @version      3.3 // バージョン更新
// @description  Firefox137向け：Equalizer Nomal
// @author       Gemini
// @match        https://*youtube.com/*
// @grant        none
// ==/UserScript==

(async function() {
    'use strict';

    // --- 設定変数 ---
    const PRIME_DECODE_WAIT_SEC = 2.00; // デコード初期化のための待機時間（秒）
    const TAP_AUDIO_DURATION_SEC = 3.00; // 音声データ取得時間（秒）
    const TARGET_SAMPLE_RATE = 96000; // 目標サンプルレート (Hz)

    // LocalEnhance 処理用設定
    const LE_CHUNK_SIZE = 1024; // チャンクサイズ (チャンネルごと)
    const LE_RMS_THRESHOLD = 0.02; // RMS閾値

    // 「弾性波動」関連定数
    const EW_VELOCITY = 340.29; // 音速 (m/s)
    // NOTE: 既存仕様によりtoFixed(2)を使用
    const EW_WAVELENGTH_FACTOR = +(EW_VELOCITY / 440).toFixed(2);
    const EW_ENHANCEMENT_FACTOR = +(Math.sin(Math.PI / EW_WAVELENGTH_FACTOR) * 0.5).toFixed(2);

    // 「Gamma式補完」関連定数
    const GAMMA_L = 1.00;
    const GAMMA_T = 1.00;
    const GAMMA_MU0 = 1.00;
    const GAMMA_KAPPA = 1.00;
    const GAMMA_ALPHAP = 1.00;
    const GAMMA_THRESHOLD = 0.5;

    // --- 一般的な音質処理用設定（模擬Dolby拡張） ---

    // 簡易疑似サラウンド設定例 (モノラルソースからのステレオ化に使用)
    const SURROUND_DELAY_LEFT_MS = 0;
    const SURROUND_DELAY_RIGHT_MS = 35;
    const SURROUND_GAIN_LEFT = 1.0;
    const SURROUND_GAIN_RIGHT = 0.8;

    // Mid/Side 処理設定
    const MID_GAIN_DB = 0; // dB
    const SIDE_GAIN_DB = 3; // dB

    // 周波数応答補正EQ群 設定
    const CORRECTION_EQ_SETTINGS = [
        { type: 'peaking', freq: 80, gain: 2, Q: 1.0 },
        { type: 'peaking', freq: 250, gain: -1.5, Q: 1.0 },
        { type: 'peaking', freq: 1000, gain: 1, Q: 1.5 },
        { type: 'peaking', freq: 4000, gain: 1.5, Q: 2.0 },
        { type: 'highshelf', freq: 15000, gain: 2 },
    ];

    // 20Hzまでの低音再現 (低域強調EQ)
    const EQ_LOWSHELF_FREQ = 30; // Hz
    const EQ_LOWSHELF_GAIN = 7; // dB

     // 高域強調EQ
    const EQ_HIGHSHELF_FREQ = 8000; // Hz
    const EQ_HIGHSHELF_GAIN = 3; // dB

     // ダイアログ強調EQ
    const EQ_DIALOG_FREQ = 2000; // Hz
    const EQ_DIALOG_GAIN = 4; // dB
    const EQ_DIALOG_Q = 2.0; // Q値

    // ダイナミックレンジ拡張/カバー
    // 低レベル信号ブースト用コンプレッサー (簡易エキスパンダー風) 設定
    const LOW_LEVEL_BOOST_THRESHOLD = -35; // dB
    const LOW_LEVEL_BOOST_RATIO = 1.0; // Web Audio API標準では1以上
    const LOW_LEVEL_BOOST_ATTACK = 0.01; // seconds
    const LOW_LEVEL_BOOST_RELEASE = 0.2; // seconds
    const LOW_LEVEL_BOOST_KNEE = 10; // dB

    // 簡易ノイズゲート/エキスパンダー風コンプレッサー 設定
    const NOISE_GATE_THRESHOLD = -50; // dB
    const NOISE_GATE_RATIO = 30; // :1
    const NOISE_GATE_ATTACK = 0.001; // seconds
    const NOISE_GATE_RELEASE = 0.15; // seconds
    const NOISE_GATE_KNEE = 2; // dB

    // ConvolverNode 用の「弾性響き」風IR設定
    const ELASTIC_CONVOLVER_IR_DURATION_SEC = 0.2; // IRの長さ (秒)
    const ELASTIC_CONVOLVER_DECAY_RATE = 0.8; // 響きの減衰率 (0-1)
    const ELASTIC_CONVOLVER_PULSE_INTERVAL_MS = 10; // パルス間の間隔 (ミリ秒)

     // ConvolverNode 用の「粗面散乱」IR設定
    const SCATTERING_ROUGHNESS_SIGMA = 0.005;    // 表面粗さの標準偏差 (m)
    const SCATTERING_CORRELATION_LENGTH = 0.02;  // 相関長 (m)
     // NOTE: 提供された generateScatteringIR は簡易的なガウス窓付きホワイトノイズを使用しています。
     // 物理モデルに基づく厳密なIR生成ではありません。

    // サブハーモニック合成設定
    const SUBHARMONIC_RATIO = 2;      // n分の1補間 (例: 2 => 1/2)
    const SUBHARMONIC_GAIN = 0.2;     // サブハーモニックの音量
     // NOTE: サブハーモニックの周波数は基本周波数推定に基づいて動的に設定されます。
     // 推定精度は estimateFundamental 関数に依存します。

    // 簡易リバーブ設定
    const REVERB_DELAY_SEC = 0.05; // ディレイ時間 (秒)
    const REVERB_FEEDBACK = 0.4; // フィードバック量 (0-1)
    const REVERB_WET_GAIN = 0.2; // リバーブ信号のミックス量 (0-1)

    // 全体のダイナミクス調整用コンプレッサー 設定
    const MAIN_COMPRESSOR_THRESHOLD = -12; // dB
    const MAIN_COMPRESSOR_RATIO = 6; // :1
    const MAIN_COMPRESSOR_ATTACK = 0.01; // seconds
    const MAIN_COMPRESSOR_RELEASE = 0.4; // seconds
    const MAIN_COMPRESSOR_KNEE = 15; // dB

    // ラウドネス調整用ゲイン
    const FINAL_GAIN = 1.0; // ゲイン (1.0 = 0dB)


    // dB値をゲイン値に変換するヘルパー関数
    function dBToGain(db) {
        return Math.pow(10, db / 20);
    }

    // シンプルな「弾性響き」風IR (AudioBuffer) を生成する関数
    function generateElasticImpulseResponse(audioContext, duration, decayRate, pulseIntervalMs) {
        const sampleRate = audioContext.sampleRate;
        const length = Math.round(duration * sampleRate);
        const buffer = audioContext.createBuffer(1, length, sampleRate); // モノラルIR
        const data = buffer.getChannelData(0);
        const pulseIntervalSamples = Math.max(1, Math.floor(sampleRate * pulseIntervalMs / 1000));

        if (length <= 0) return buffer;

        data[0] = 1.0; // 最初のパルス

        for (let i = pulseIntervalSamples; i < length; i += pulseIntervalSamples) {
            const time = i / sampleRate;
            const amplitude = Math.pow(decayRate, time) * (1.0 + Math.random() * 0.2 - 0.1); // 指数関数的減衰 + 少しのランダム揺らぎ
            if (i < length) { // バッファ範囲チェック
                data[i] = amplitude;
            }
        }
        return buffer;
    }

    // 「粗面散乱」IR (AudioBuffer) を生成する関数 (提供されたロジックをそのまま使用)
     function generateScatteringIR(audioCtx, sigma, Lc) {
        const fs = audioCtx.sampleRate;
        const length = Math.round(0.1 * fs); // 0.1秒固定
        const buffer = audioCtx.createBuffer(1, length, fs);
        const data = buffer.getChannelData(0);

        // 単純化: ガウス窓付きホワイトノイズ
        for (let i = 0; i < length; i++) {
            const t = i / fs;
            const env = Math.exp(-t * 5);    // 衰減 (固定の衰減率)
            data[i] = (Math.random()*2 - 1) * env; // ランダム値を使用
        }
        return buffer;
    }

    // サブハーモニック合成ノードを作成する関数 (提供されたロジックをそのまま使用)
     function createSubharmonicNode(audioCtx, ratio, gainValue) {
        const osc = audioCtx.createOscillator();
        osc.type = 'sine';
        const gain = audioCtx.createGain();
        gain.gain.value = gainValue;
        osc.connect(gain);
        return { osc, gain };
    }


    // 基本周波数を推定する関数 (提供されたロジックをそのまま使用)
    // NOTE: この関数は AnalyserNode を入力とし、自己相関法で基本周波数を推定します。
    // 精度はデータ長や信号に依存します。OfflineAudioContext内では直接使用できません。
    function estimateFundamental(analyser) {
        if (!analyser) {
             console.warn("AnalyserNode not provided for fundamental frequency estimation.");
             return 0; // 推定失敗
        }
        const data = new Float32Array(analyser.fftSize);
        analyser.getFloatTimeDomainData(data);
        // Auto-correlation 法で f0 を推定 (簡易)
        let bestOffset = -1; // 0 から開始すると信号がゼロの場合に bestCorr がゼロになる
        let bestCorr = -Infinity; // 負の無限大で初期化
        const sampleRate = analyser.context.sampleRate;
        const maxOffset = Math.min(Math.floor(sampleRate / 20), data.length / 2); // 最大オフセットを制限 (最低20Hzまで)
        const minOffset = Math.max(2, Math.floor(sampleRate / 500)); // 最小オフセットを制限 (最高500Hzまで)

        for (let off = minOffset; off < maxOffset; off++) {
            let corr = 0;
            for (let i = 0; i < data.length - off; i++) {
                corr += data[i] * data[i + off];
            }
            // 自己相関値の正規化は行っていません (簡易)
            if (corr > bestCorr) { bestCorr = corr; bestOffset = off; }
        }
        // bestCorrがある程度大きい場合のみ有効な結果とみなす (簡易)
        if (bestCorr > 0.1 * data.length) { // 適当な閾値
            const fundamental = sampleRate / bestOffset;
             console.log(`Estimated fundamental frequency: ${fundamental.toFixed(2)} Hz (Offset: ${bestOffset})`);
            return fundamental;
        } else {
             console.log("Fundamental frequency estimation failed (correlation too low).");
             return 0; // 推定失敗
        }
    }


     // IR生成をUserScript起動時に一度だけ実行しキャッシュ
    let elasticImpulseResponseBuffer = null;
     // scattering IR も起動時に生成しキャッシュ (OfflineContext内で使用)
    let scatteringImpulseResponseBuffer = null;

    // 起動時に仮のAudioContextを作成してIRを生成
    try {
         // IR生成時のサンプルレートは、OfflineContextのTARGET_SAMPLE_RATEに合わせる
         const tempCtxForIR = new OfflineAudioContext(1, 1, TARGET_SAMPLE_RATE);
         elasticImpulseResponseBuffer = generateElasticImpulseResponse(tempCtxForIR, ELASTIC_CONVOLVER_IR_DURATION_SEC, ELASTIC_CONVOLVER_DECAY_RATE, ELASTIC_CONVOLVER_PULSE_INTERVAL_MS);
         scatteringImpulseResponseBuffer = generateScatteringIR(tempCtxForIR, SCATTERING_ROUGHNESS_SIGMA, SCATTERING_CORRELATION_LENGTH);
         tempCtxForIR.close().catch(() => {});
         console.log("Impulse Response buffers generated and cached.");
    } catch (e) {
         console.error("Failed to generate Impulse Response buffers:", e);
         elasticImpulseResponseBuffer = null;
         scatteringImpulseResponseBuffer = null;
    }


    // --- AudioWorkletProcessor definition ---
    // PCMデータをそのまま出力し、メインスレッドに転送するWorklet
    const processorCode = `
    class PCMProcessor extends AudioWorkletProcessor {
        process(inputs, outputs) {
            const input = inputs[0];
            const output = outputs[0];
            const channelCount = input ? input.length : 0;

            if (channelCount > 0) {
                const channelData = [];
                for(let i = 0; i < channelCount; i++) {
                    if (input[i]) {
                        channelData.push(new Float32Array(input[i])); // コピーして送信
                        if (output[i]) {
                           output[i].set(input[i]);
                        }
                    } else {
                         channelData.push(new Float32Array(0));
                    }
                }
                if (channelData.length > 0) {
                   this.port.postMessage(channelData);
                }
            }
            return true;
        }
    }
    registerProcessor('pcm-processor', PCMProcessor);
    `;

    const blob = new Blob([processorCode], { type: 'application/javascript' });
    let moduleURL = URL.createObjectURL(blob);

    const activeAudioContexts = [];


    // メディア要素（audioまたはvideo）が利用可能になるまで待機
    /*
    const waitForMedia = () => new Promise(resolve => {
        console.log("Waiting for media element...");
        const tryFind = () => {
            const m = document.querySelector('audio, video');
            if (m && m.readyState >= 2) {
                 console.log("Media element found and ready.");
                 resolve(m);
            } else {
                 setTimeout(tryFind, 500);
            }
        };
        tryFind();
    });
    */

    // メディアのデコードを初期化するために少し再生を試みる
    const primeDecode = async (media, waitSec = PRIME_DECODE_WAIT_SEC) => {
        console.log(`Starting prime decode for ${waitSec} seconds...`);
        media.muted = true;
        try {
            await media.play();
        } catch (e) {
            console.warn("Media play for prime decode failed:", e);
        }
        await new Promise(r => setTimeout(r, waitSec * 1000));
        console.log("Prime decode finished.");
    };

    // AudioWorkletを使用して指定時間だけ音声データを取得 (ステレオ対応)
    // 基本周波数推定のためにAnalyserNodeを接続し、データを取得後推定を実行
    async function tapAudio(media, durationSec = TAP_AUDIO_DURATION_SEC) {
        console.log("Starting tapAudio...");
        const ctx = new AudioContext();
        activeAudioContexts.push(ctx);

        try {
            await ctx.audioWorklet.addModule(moduleURL);
            console.log("AudioWorklet module added.");
        } catch (e) {
            console.error("Failed to add AudioWorklet module:", e);
            ctx.close().catch(() => {});
            activeAudioContexts.splice(activeAudioContexts.indexOf(ctx), 1);
            throw e;
        }

        const source = ctx.createMediaElementSource(media);
        const workletNode = new AudioWorkletNode(ctx, 'pcm-processor', {
             outputChannelCount:[source.channelCount]
        });
         console.log(`Audio source channels: ${source.channelCount}`);

        // 基本周波数推定用のAnalyserNodeを追加
        const analyserNode = ctx.createAnalyser();
        analyserNode.fftSize = 2048; // FFTサイズを設定 (推定精度に影響)

        const chunks = [];
        workletNode.port.onmessage = e => {
            if (e.data && Array.isArray(e.data) && e.data.length > 0 && Array.isArray(e.data[0])) {
                chunks.push(e.data);
            } else {
                console.warn("Received unexpected data format from Worklet:", e.data);
            }
        };

        source.connect(analyserNode); // ソースをAnalyserに接続
        analyserNode.connect(workletNode); // Analyserの後にWorkletを接続
        workletNode.connect(ctx.destination);

        console.log(`Collecting audio data for ${durationSec} seconds...`);
        await new Promise(r => setTimeout(r, durationSec * 1000));
        console.log("Finished collecting audio data.");

        source.disconnect(); // 全ての接続を切断
        analyserNode.disconnect();
        workletNode.disconnect();

        // データを取得した後、AnalyserNodeを使って基本周波数を推定
         console.log("Estimating fundamental frequency...");
        const fundamentalFrequency = estimateFundamental(analyserNode);
         console.log(`Estimated f0: ${fundamentalFrequency.toFixed(2)} Hz`);


        ctx.close().catch(e => console.error("tapAudio context close failed:", e));
        activeAudioContexts.splice(activeAudioContexts.indexOf(ctx), 1);

        const channelCount = chunks.length > 0 && chunks[0] && Array.isArray(chunks[0]) ? chunks[0].length : 0;
        const pcmByChannel = [];

        if (channelCount > 0) {
            const channelLengths = new Array(channelCount).fill(0);
            for(const chunk of chunks) {
                 if (Array.isArray(chunk)) {
                     for(let i = 0; i < channelCount; i++) {
                         if (chunk[i] && Array.isArray(chunk[i])) channelLengths[i] += chunk[i].length;
                     }
                 }
            }
            for(let i = 0; i < channelCount; i++) {
                 pcmByChannel.push(new Float32Array(channelLengths[i]));
            }

            const offsets = new Array(channelCount).fill(0);
            for(const chunk of chunks) {
                 if (Array.isArray(chunk)) {
                     for(let i = 0; i < channelCount; i++) {
                         if (chunk[i] && Array.isArray(chunk[i])) {
                            pcmByChannel[i].set(chunk[i], offsets[i]);
                            offsets[i] += chunk[i].length;
                         }
                     }
                 }
            }
        }
         console.log(`Collected ${channelCount} channels of PCM data. Total samples (per channel): ${pcmByChannel.length > 0 ? pcmByChannel[0].length : 0}`);

        return {pcmByChannel, sampleRate:ctx.sampleRate, channelCount, fundamentalFrequency};
    }


    // 独自の音質エンハンスメント処理（ローカル領域）と全体的な模擬Dolby処理を統合し、
    // 一度のOfflineAudioContextレンダリングで実行 (ステレオ対応)
    // ローカル処理はメインスレッドで行われます。spliceによるバッファ改変の問題は維持されます。
    async function processAudio(pcmByChannel, origRate, channelCount, fundamentalFrequency) {
         console.log("Starting processAudio (main thread processing + Offline rendering)...");

        // 1. メインスレッドでローカルエンハンスメントの大部分（Gamma式サンプル挿入まで）を実行
         console.log("Applying local enhancement (main thread)...");
        const processedLocalPcmByChannel = [];
        const localInfoByChannel = [];

        for(let ch = 0; ch < channelCount; ch++) {
            const pcm = pcmByChannel[ch];
            if (!pcm || pcm.length === 0) {
                 processedLocalPcmByChannel.push(new Float32Array(0));
                 localInfoByChannel.push(null);
                 continue;
            }

            // 元の localEnhance ロジックをチャンネルごとに適用 (toFixed(2), splice 等もユーザー要望により維持)
            const chunkSize = LE_CHUNK_SIZE;
            const total = Math.floor(pcm.length / chunkSize);
            // 処理に必要な最小限のデータ長をチェック
            if (total < 3 || pcm.length < chunkSize * 3 + 256 * 2 + 2) {
                 console.warn(`Channel ${ch}: PCM data too short (${pcm.length} samples) for local enhancement. Skipping local processing for this channel.`);
                 processedLocalPcmByChannel.push(new Float32Array(pcm));
                 localInfoByChannel.push(null);
                 continue;
            }


            let amps = new Float32Array(total).fill(1 / Math.sqrt(total));

            for(let i=0; i<total; i++){
                const c = pcm.subarray(i * chunkSize, (i + 1) * chunkSize);
                const rms = Math.sqrt(c.reduce((s, v) => s + v * v, 0) / chunkSize);
                if (rms < LE_RMS_THRESHOLD) amps[i] = -1;
            }

            const mean = amps.reduce((a, b) => a + b, 0) / total;
            amps = amps.map(a => 2 * mean - a);

            const maxIdx = amps.map((a, i) => ({ p: a * a, i })).sort((a, b) => b.p - a.p)[0].i;
            const base = maxIdx * chunkSize;

            const preStart = Math.max(0, base - 256);
            const postEnd = Math.min(pcm.length, base + chunkSize + 256);
            const pre = pcm.subarray(preStart, base);
            const target = pcm.subarray(base, base + chunkSize);
            const post = pcm.subarray(base + chunkSize, postEnd);

            const vel = EW_VELOCITY;
            const wl = EW_WAVELENGTH_FACTOR;
            const ef = EW_ENHANCEMENT_FACTOR;

            const elPre = new Float32Array(pre.length);
            const elPost = new Float32Array(post.length);
             // target[i%chunkSize] が target.length を超えないように安全に参照
            for(let i = 0; i < pre.length; i++){
                const w = +((i / pre.length) * ef).toFixed(2); // 既存仕様によりtoFixed(2)
                 const targetIdx = i % chunkSize;
                 const targetValue = targetIdx < target.length ? target[targetIdx] : 0;
                 elPre[i] = +(pre[i] * (1 - w) + targetValue * w).toFixed(2); // 既存仕様によりtoFixed(2)
            }
            for(let i = 0; i < post.length; i++){
                const w = +((i / post.length) * ef).toFixed(2); // 既存仕様によりtoFixed(2)
                 const targetIdx = i % chunkSize;
                 const targetValue = targetIdx < target.length ? target[targetIdx] : 0;
                elPost[i] = +(post[i] * (1 - w) + targetValue * w).toFixed(2); // 既存仕様によりtoFixed(2)
            }

            const idxs = [base - 2, base - 1, base + chunkSize, base + chunkSize + 1];
            const safeIdxs = idxs.map(i => Math.min(Math.max(i, 0), pcm.length - 1)); // 範囲チェック
            const y = safeIdxs.map(i => pcm[i]);

            const fill = new Float32Array(chunkSize);
            for(let j = 0; j < chunkSize; j++) {
                fill[j] = cubicInterpolate(y[0], y[1], y[2], y[3], j / (chunkSize - 1)); // 既存仕様によりtoFixed(2)内包
            }

             const regionStart = (maxIdx - 1) * chunkSize;
             const regionEnd = (maxIdx + 2) * chunkSize;
             const safeRegionStart = Math.max(0, regionStart);
             const safeRegionEnd = Math.min(pcm.length, regionEnd);

            const region = pcm.subarray(safeRegionStart, safeRegionEnd);

            const avg = region.reduce((s, v) => s + v, 0) / (region.length || 1);
            const gain = +((1 / (Math.abs(avg) + 1e-6)).toFixed(2)); // 既存仕様によりtoFixed(2)

            const amplified = new Float32Array(chunkSize);
            for(let i = 0; i < chunkSize; i++) {
                amplified[i] = +(target[i] * gain).toFixed(2); // 既存仕様によりtoFixed(2)
            }

            // === Gamma processing ===
            const L = GAMMA_L;
            const T = GAMMA_T;
            const mu0 = GAMMA_MU0;
            const kappa = GAMMA_KAPPA;
            const alphaP = GAMMA_ALPHAP;
            const gammaThreshold = GAMMA_THRESHOLD;

            let full = [...elPre, ...fill, ...amplified, ...fill, ...elPost];
            // NOTE: spliceによる可変長バッファ改変はタイムラインズレや位相歪みの原因となりますが、ユーザー要望により既存仕様を維持します。
            for(let i = 1; i < full.length; i++){
              const mu_i = Math.abs(full[i]);
              let gamma = 0;
              if (mu_i > 1e-9) {
                 const logArg = Math.abs((Math.sqrt(mu0) - Math.sqrt(mu_i)) / (Math.sqrt(mu0) + Math.sqrt(mu_i)));
                 if (logArg > 1e-9) {
                    gamma = -1 / L * Math.sqrt(T / mu_i) * Math.log(logArg) * (1 + kappa * alphaP / (L * L));
                 }
              }
              const g = +gamma.toFixed(2); // 既存仕様によりtoFixed(2)

              if(Math.abs(g) > gammaThreshold){
                const extra = +(((full[i - 1] + full[i]) / 2).toFixed(2)); // 既存仕様によりtoFixed(2)
                full.splice(i, 0, extra);
                i++;
              }
            }
            processedLocalPcmByChannel.push(new Float32Array(full));

             const localStartIndex = preStart;
             localInfoByChannel.push({
                 startIndex: localStartIndex,
                 processedLength: full.length
             });
        }
         console.log("Local enhancement processing complete for all channels.");


        // 2. 元のPCMデータに、ローカル処理済みPCMデータを上書き合成した最終PCMバッファを作成 (メインスレッド)
         console.log("Synthesizing final PCM buffer (main thread)...");
        const originalTotalLength = pcmByChannel.length > 0 ? pcmByChannel[0].length : 0;
        const finalPcmByChannel = [];

        for(let ch = 0; ch < channelCount; ch++) {
             const originalPcm = pcmByChannel[ch];
             const processedLocalPcm = processedLocalPcmByChannel[ch];
             const localInfo = localInfoByChannel[ch];

             if (!originalPcm) {
                 finalPcmByChannel.push(new Float32Array(0));
                 continue;
             }

             const finalPcm = new Float32Array(originalPcm);

             if (localInfo && processedLocalPcm && processedLocalPcm.length > 0) {
                 const startIndex = localInfo.startIndex;
                 const copyLength = Math.min(processedLocalPcm.length, finalPcm.length - startIndex);

                 if (startIndex < finalPcm.length && copyLength > 0) {
                    const sourceData = processedLocalPcm.subarray(0, copyLength);
                    finalPcm.set(sourceData, startIndex);
                 }
             }
             finalPcmByChannel.push(finalPcm);
        }
        console.log("Final PCM buffer synthesized.");


        // 3. 一度のOfflineAudioContextレンダリングで、合成済み最終PCMデータに対して
        //    リサンプリング (origRate -> TARGET_SAMPLE_RATE) と模擬Dolby処理を実行

        // 出力バッファ長を計算 (元の長さ * リサンプル比率)
        const finalRenderedLength = Math.floor(originalTotalLength * TARGET_SAMPLE_RATE / origRate);
        const offlineContext = new OfflineAudioContext(channelCount, finalRenderedLength, TARGET_SAMPLE_RATE);
        activeAudioContexts.push(offlineContext);
         console.log(`Created OfflineAudioContext (channels: ${channelCount}, length: ${finalRenderedLength}, sampleRate: ${TARGET_SAMPLE_RATE}).`);


        // 最終合成済みPCMデータをOfflineContextのBufferに読み込み
        const finalPcmBuffer = offlineContext.createBuffer(channelCount, originalTotalLength, origRate);
         for(let i = 0; i < channelCount; i++) {
             if (finalPcmByChannel[i]) {
                 finalPcmBuffer.copyToChannel(finalPcmByChannel[i], i);
             }
         }
         console.log("Loaded final PCM into OfflineContext buffer.");

        const sourceInOffline = offlineContext.createBufferSource();
        sourceInOffline.buffer = finalPcmBuffer;


        // --- OfflineAudioContext内の模擬Dolby処理ノードグラフ ---
        // 設定変数を使用し、OfflineContext用のノードを作成・接続

        // 簡易疑似サラウンド処理 (OfflineContext用)
        const splitterSimSurround_off = offlineContext.createChannelSplitter(channelCount);
        const mergerSimSurround_off = offlineContext.createChannelMerger(channelCount);

        const delayLeftSimSurround_off = offlineContext.createDelay(1.0);
        delayLeftSimSurround_off.delayTime.setValueAtTime(SURROUND_DELAY_LEFT_MS / 1000, offlineContext.currentTime);
        const gainLeftSimSurround_off = offlineContext.createGain();
        gainLeftSimSurround_off.gain.setValueAtTime(SURROUND_GAIN_LEFT, offlineContext.currentTime);

        const delayRightSimSurround_off = offlineContext.createDelay(1.0);
        delayRightSimSurround_off.delayTime.setValueAtTime(SURROUND_DELAY_RIGHT_MS / 1000, offlineContext.currentTime);
        const gainRightSimSurround_off = offlineContext.createGain();
        gainRightSimSurround_off.gain.setValueAtTime(SURROUND_GAIN_RIGHT, offlineContext.currentTime);


        // Mid/Side 処理ノード群 (OfflineContext用)
        const splitterMS = offlineContext.createChannelSplitter(channelCount);
        const mergerMS = offlineContext.createChannelMerger(channelCount);

        const midGain = offlineContext.createGain(); midGain.gain.setValueAtTime(0.5, offlineContext.currentTime);
        const sideGain = offlineContext.createGain(); sideGain.gain.setValueAtTime(0.5, offlineContext.currentTime);
        const negSideGain = offlineContext.createGain(); negSideGain.gain.setValueAtTime(-0.5, offlineContext.currentTime);

        const midProcessGain = offlineContext.createGain();
        midProcessGain.gain.setValueAtTime(dBToGain(MID_GAIN_DB), offlineContext.currentTime);

        const sideProcessGain = offlineContext.createGain();
        sideProcessGain.gain.setValueAtAtTime(dBToGain(SIDE_GAIN_DB), offlineContext.currentTime); // Typo修正 setValueAtTime


        const gainOne = offlineContext.createGain(); gainOne.gain.setValueAtTime(1.0, offlineContext.currentTime);
        const gainNegOne = offlineContext.createGain(); gainNegOne.gain.setValueAtTime(-1.0, offlineContext.currentTime);


        // 周波数応答補正EQ群 (OfflineContext用)
        const correctionEQs_off = CORRECTION_EQ_SETTINGS.map(settings => {
            const eq = offlineContext.createBiquadFilter();
            eq.type = settings.type;
            eq.frequency.setValueAtTime(settings.freq, offlineContext.currentTime);
            eq.gain.setValueAtTime(settings.gain, offlineContext.currentTime);
            if (settings.Q) eq.Q.setValueAtTime(settings.Q, offlineContext.currentTime);
            return eq;
        });

        // 低域強調EQ (OfflineContext用)
        const eqLow_off = offlineContext.createBiquadFilter();
        eqLow_off.type = 'lowshelf';
        eqLow_off.frequency.setValueAtTime(EQ_LOWSHELF_FREQ, offlineContext.currentTime);
        eqLow_off.gain.setValueAtTime(EQ_LOWSHELF_GAIN, offlineContext.currentTime);

         // 高域強調EQ (OfflineContext用)
        const eqHigh_off = offlineContext.createBiquadFilter();
        eqHigh_off.type = 'highshelf';
        eqHigh_off.frequency.setValueAtTime(EQ_HIGHSHELF_FREQ, offlineContext.currentTime);
        eqHigh_off.gain.setValueAtTime(EQ_HIGHSHELF_GAIN, offlineContext.currentTime);

         // ダイアログ強調EQ (OfflineContext用)
        const eqDialog_off = offlineContext.createBiquadFilter();
        eqDialog_off.type = 'peaking';
        eqDialog_off.frequency.setValueAtTime(EQ_DIALOG_FREQ, offlineContext.currentTime);
        eqDialog_off.gain.setValueAtTime(EQ_DIALOG_GAIN, offlineContext.currentTime);
        eqDialog_off.Q.setValueAtTime(EQ_DIALOG_Q, offlineContext.currentTime);


        // 低レベル信号ブースト用コンプレッサー (OfflineContext用)
        const lowLevelBoostComp_off = offlineContext.createDynamicsCompressor();
        lowLevelBoostComp_off.threshold.setValueAtTime(LOW_LEVEL_BOOST_THRESHOLD, offlineContext.currentTime);
        lowLevelBoostComp_off.ratio.setValueAtTime(LOW_LEVEL_BOOST_RATIO, offlineContext.currentTime);
        lowLevelBoostComp_off.attack.setValueAtTime(LOW_LEVEL_BOOST_ATTACK, offlineContext.currentTime);
        lowLevelBoostComp_off.release.setValueAtTime(LOW_LEVEL_BOOST_RELEASE, offlineContext.currentTime);
        lowLevelBoostComp_off.knee.setValueAtTime(LOW_LEVEL_BOOST_KNEE, offlineContext.currentTime);
        lowLevelBoostComp_off.release.setValueAtTime(LOW_LEVEL_BOOST_RELEASE, offlineContext.currentTime); // 重複削除


        // 簡易ノイズゲート/エキスパンダー風コンプレッサー (OfflineContext用)
        const noiseGateComp_off = offlineContext.createDynamicsCompressor();
        noiseGateComp_off.threshold.setValueAtTime(NOISE_GATE_THRESHOLD, offlineContext.currentTime);
        noiseGateComp_off.ratio.setValueAtTime(NOISE_GATE_RATIO, offlineContext.currentTime);
        noiseGateComp_off.attack.setValueAtTime(NOISE_GATE_ATTACK, offlineContext.currentTime);
        noiseGateComp_off.release.setValueAtTime(NOISE_GATE_RELEASE, offlineContext.currentTime);
        noiseGateComp_off.knee.setValueAtTime(NOISE_GATE_KNEE, offlineContext.currentTime);


        // ConvolverNode (弾性響き) (OfflineContext用)
        const elasticConvolverNode = offlineContext.createConvolver();
        if (elasticImpulseResponseBuffer && elasticImpulseResponseBuffer.sampleRate === offlineContext.sampleRate) {
             elasticConvolverNode.buffer = elasticImpulseResponseBuffer;
        } else {
            console.warn(`Elastic Impulse Response buffer not available or sample rate mismatch (IR: ${elasticImpulseResponseBuffer ? elasticImpulseResponseBuffer.sampleRate : 'N/A'}, OfflineCtx: ${offlineContext.sampleRate}). Elastic Convolver will not apply effect.`);
             const dummyIr = offlineContext.createBuffer(1, 1, offlineContext.sampleRate);
             dummyIr.getChannelData(0)[0] = 0;
             elasticConvolverNode.buffer = dummyIr;
        }

         // ConvolverNode (粗面散乱) (OfflineContext用) - 新しく追加
        const scatteringConvolverNode = offlineContext.createConvolver();
         // Scattering IR 生成と設定
         try {
             const scatteringIr = generateScatteringIR(offlineContext, SCATTERING_ROUGHNESS_SIGMA, SCATTERING_CORRELATION_LENGTH);
             scatteringConvolverNode.buffer = scatteringIr;
              console.log("Scattering ConvolverNode IR generated and set.");
         } catch (e) {
             console.error("Failed to generate or set Scattering IR for ConvolverNode:", e);
             // エラー時もダミーIRを設定
             const dummyIr = offlineContext.createBuffer(1, 1, offlineContext.sampleRate);
             dummyIr.getChannelData(0)[0] = 0;
             scatteringConvolverNode.buffer = dummyIr;
         }


        // サブハーモニック合成ノード (OfflineContext用) - 新しく追加
        const { osc: subOsc, gain: subGain } = createSubharmonicNode(offlineContext, SUBHARMONIC_RATIO, SUBHARMONIC_GAIN);
         // 基本周波数を設定
         if (fundamentalFrequency > 0) {
              // サブハーモニック周波数 = fundamentalFrequency / SUBHARMONIC_RATIO
              subOsc.frequency.setValueAtTime(fundamentalFrequency / SUBHARMONIC_RATIO, offlineContext.currentTime);
              subOsc.start(0); // OfflineContextの最初から開始
               console.log(`Subharmonic oscillator frequency set to ${fundamentalFrequency / SUBHARMONIC_RATIO} Hz.`);
         } else {
              console.warn("Fundamental frequency not estimated or is zero. Subharmonic oscillator will not be started.");
              // 推定失敗の場合はオシレーターを開始しない
         }


        // 簡易リバーブ処理 (フィードバックディレイ) (OfflineContext用)
        const reverbDelay_off = offlineContext.createDelay(1.0);
        reverbDelay_off.delayTime.setValueAtTime(REVERB_DELAY_SEC, offlineContext.currentTime);
        const reverbFeedbackGain_off = offlineContext.createGain();
        reverbFeedbackGain_off.gain.setValueAtTime(REVERB_FEEDBACK, offlineContext.currentTime);
        const reverbWetGain_off = offlineContext.createGain();
        reverbWetGain_off.gain.setValueAtTime(REVERB_WET_GAIN, offlineContext.currentTime);
        const reverbDryGain_off = offlineContext.createGain();
        reverbDryGain_off.gain.setValueAtTime(1.0 - REVERB_WET_GAIN, offlineContext.currentTime);

        reverbDelay_off.connect(reverbFeedbackGain_off);
        reverbFeedbackGain_off.connect(reverbDelay_off);

        const reverbMixMerger_off = offlineContext.createChannelMerger(channelCount);


        // 全体のダイナミクス調整用コンプレッサー (OfflineContext用)
        const mainCompressor_off = offlineContext.createDynamicsCompressor();
        mainCompressor_off.threshold.setValueAtTime(MAIN_COMPRESSOR_THRESHOLD, offlineContext.currentTime);
        mainCompressor_off.ratio.setValueAtTime(MAIN_COMPRESSOR_RATIO, offlineContext.currentTime);
        mainCompressor_off.attack.setValueAtTime(MAIN_COMPRESSOR_ATTACK, offlineContext.currentTime);
        mainCompressor_off.release.setValueAtTime(MAIN_COMPRESSOR_RELEASE, offlineContext.currentTime);
        mainCompressor_off.knee.setValueAtTime(MAIN_COMPRESSOR_KNEE, offlineContext.currentTime);

        // ラウドネス調整用ゲイン (OfflineContext用)
        const finalGainNode_off = offlineContext.createGain();
        finalGainNode_off.gain.setValueAtTime(FINAL_GAIN, offlineContext.currentTime);


        // --- OfflineAudioContext内でのノード接続 ---

        // Source -> 簡易疑似サラウンド処理
        let lastNode_off = sourceInOffline;
        lastNode_off.connect(splitterSimSurround_off);

        // 簡易疑似サラウンド処理接続
        if (channelCount > 0 && splitterSimSurround_off.numberOfOutputs > 0) {
             splitterSimSurround_off.connect(delayLeftSimSurround_off, 0);
             delayLeftSimSurround_off.connect(gainLeftSimSurround_off);
             gainLeftSimSurround_off.connect(mergerSimSurround_off, 0, 0);

             if (channelCount > 1 && splitterSimSurround_off.numberOfOutputs > 1) {
                splitterSimSurround_off.connect(delayRightSimSurround_off, 1);
                delayRightSimSurround_off.connect(gainRightSimSurround_off);
                gainRightSimSurround_off.connect(mergerSimSurround_off, 0, 1);
             } else { // モノラルソースの場合
                 splitterSimSurround_off.connect(delayRightSimSurround_off, 0);
                 delayRightSimSurround_off.connect(gainRightSimSurround_off);
                 gainRightSimSurround_off.connect(mergerSimSurround_off, 0, 1);
             }
        }
        lastNode_off = mergerSimSurround_off;


        // 簡易疑似サラウンドの出力 -> Mid/Side分解 -> Mid処理 -> Side処理 -> Mid/Side合成
        lastNode_off.connect(splitterMS);

        splitterMS.connect(midGain, 0);
        splitterMS.connect(midGain, 1);

        splitterMS.connect(sideGain, 0);
        splitterMS.connect(negSideGain, 1);
        negSideGain.connect(sideGain);

        midGain.connect(midProcessGain);
        sideGain.connect(sideProcessGain);

        midProcessGain.connect(mergerMS, 0, 0);
        sideProcessGain.connect(mergerMS, 0, 0);

        midProcessGain.connect(mergerMS, 0, 1);
        sideProcessGain.connect(gainNegOne, 0);
        gainNegOne.connect(mergerMS, 0, 1);

        lastNode_off = mergerMS;


        // Mid/Side合成の出力 -> 周波数応答補正EQ群 -> 低域EQ -> 高域EQ -> ダイアログEQ -> 低レベルブーストコンプ -> ノイズゲート風コンプ
        correctionEQs_off.forEach(eq => {
            lastNode_off.connect(eq);
            lastNode_off = eq;
        });

        lastNode_off.connect(eqLow_off);
        eqLow_off.connect(eqHigh_off);
        eqHigh_off.connect(eqDialog_off);
        lastNode_off = eqDialog_off;


        lastNode_off.connect(lowLevelBoostComp_off);
        lowLevelBoostComp_off.connect(noiseGateComp_off);
        lastNode_off = noiseGateComp_off;


        // ConvolverNode (弾性響き) 接続
        lastNode_off.connect(elasticConvolverNode);
        lastNode_off = elasticConvolverNode;

         // ConvolverNode (粗面散乱) 接続 - 新しく追加
        lastNode_off.connect(scatteringConvolverNode);
        lastNode_off = scatteringConvolverNode;


        // リバーブ接続: 粗面散乱Convolver の出力をドライ信号とウェット信号に分岐
        lastNode_off.connect(reverbDryGain_off);

        const convolverOutputMixer_off = offlineContext.createChannelMerger(1);
        lastNode_off.connect(convolverOutputMixer_off);
        convolverOutputMixer_off.connect(reverbDelay_off);

        reverbDryGain_off.connect(reverbMixMerger_off, 0, 0);
        reverbDryGain_off.connect(reverbMixMerger_off, 0, 1);

        reverbDelay_off.connect(reverbWetGain_off);
        reverbWetGain_off.connect(reverbMixMerger_off, 0, 0);
        reverbWetGain_off.connect(reverbMixMerger_off, 0, 1);

        lastNode_off = reverbMixMerger_off;

         // サブハーモニック信号をリバーブミックスに加算 - 新しく追加
         subGain.connect(lastNode_off); // subGain の出力をリバーブミックスに接続

        // リバーブミックス + サブハーモニック -> メインコンプレッサー
        lastNode_off.connect(mainCompressor_off); // lastNode_off は reverbMixMerger_off に更新済み

        // メインコンプレッサー -> ラウドネス調整Gain -> Destination
        mainCompressor_off.connect(finalGainNode_off);
        finalGainNode_off.connect(offlineContext.destination);


        // レンダリング開始
        console.log("Starting OfflineAudioContext rendering...");
        sourceInOffline.start(0);
        const renderedBuffer = await offlineContext.startRendering();
        console.log("OfflineAudioContext rendering finished.");

        // Contextをクローズ
        offlineContext.close().catch(e => console.error("Offline context close failed:", e));
        activeAudioContexts.splice(activeAudioContexts.indexOf(offlineContext), 1);

        return renderedBuffer;
    }


    // 最終レンダリング結果を再生
    async function mixAndPlay(finalRenderedBuffer){
         console.log("Starting playback...");
        const ctx = new AudioContext({ sampleRate: TARGET_SAMPLE_RATE });
        activeAudioContexts.push(ctx);

        const sourceNode = ctx.createBufferSource();
        sourceNode.buffer = finalRenderedBuffer;

        sourceNode.onended = () => {
            console.log("Audio playback ended. Closing context and revoking URL.");
            ctx.close().catch(e => console.error("Playback context close failed:", e));
             activeAudioContexts.splice(activeAudioContexts.indexOf(ctx), 1);
            if (moduleURL) {
                URL.revokeObjectURL(moduleURL);
                moduleURL = null;
            }
        };

        sourceNode.connect(ctx.destination);

        sourceNode.start();
         console.log("Playback started.");
    }

    // --- リソース解放処理 ---
    window.addEventListener('unload', () => {
        console.log("Page unloading. Closing active contexts.");
        activeAudioContexts.forEach(ctx => {
            if (ctx && ctx.state !== 'closed') {
                ctx.close().catch(() => {});
            }
        });
        activeAudioContexts.length = 0;

        if (moduleURL) {
            URL.revokeObjectURL(moduleURL);
            moduleURL = null;
        }
         console.log("Unload cleanup complete.");
    });

     // エラー発生時のクリーンアップ
     const cleanupOnError = (e) => {
         console.error("An error occurred during execution:", e);
         activeAudioContexts.forEach(ctx => {
             if (ctx && ctx.state !== 'closed') {
                 ctx.close().catch(() => {});
             }
         });
         activeAudioContexts.length = 0;
         if (moduleURL) {
             URL.revokeObjectURL(moduleURL);
             moduleURL = null;
         }
         console.log("Error cleanup complete.");
     };


    // UserScript実行開始
    try {
        const media = await waitForMedia();
        await primeDecode(media, PRIME_DECODE_WAIT_SEC);
         // tapAudioで基本周波数推定も行う
        const {pcmByChannel, sampleRate, channelCount, fundamentalFrequency} = await tapAudio(media, TAP_AUDIO_DURATION_SEC);
        // processAudioに基本周波数推定結果を渡す
        const finalRenderedBuffer = await processAudio(pcmByChannel, sampleRate, channelCount, fundamentalFrequency);
        await mixAndPlay(finalRenderedBuffer);

    } catch (e) {
        cleanupOnError(e);
    }

})();

